use std::{
    collections::{HashMap, HashSet, VecDeque},
    hash::Hash,
    sync::Mutex,
};

use itertools::Itertools;
use lazy_static::lazy_static;
use regex::Regex;

use crate::util;

const TOTAL_TIME: usize = 30;

#[derive(Debug, Clone)]
pub struct Valve {
    pub flow_rate: usize,
    pub tunnels: Vec<usize>,
}

pub fn parse_valves(input: &String) -> Vec<Valve> {
    let indices: HashMap<String, usize> = input
        .lines()
        .map(|l| l.split(' ').nth(1).unwrap())
        .sorted()
        .enumerate()
        .map(|(i, n)| (n.to_owned(), i))
        .collect();

    let mut valves = HashMap::new();

    let re =
        Regex::new(r"^Valve (\w+) has flow rate=(\d+); tunnels? leads? to valves? (.*)$").unwrap();
    for line in input.lines() {
        let captures = re
            .captures(line)
            .expect(&format!("No captures for {}", line));
        let name = captures.get(1).unwrap().as_str().to_owned();
        let flow_rate = captures.get(2).unwrap().as_str().parse().unwrap();
        let tunnels = captures
            .get(3)
            .unwrap()
            .as_str()
            .split(", ")
            .map(|s| indices[s])
            .collect();
        let valve = Valve { flow_rate, tunnels };
        valves.insert(name, valve);
    }

    valves
        .iter()
        .map(|(name, v)| (indices[name], v))
        .sorted_by_key(|(i, _)| *i)
        .map(|(_, v)| v.clone())
        .collect()
}

pub fn calc_dists(valves: &Vec<Valve>) -> Vec<Vec<usize>> {
    let n = valves.len();
    let mut result = vec![vec![usize::MAX; n]; n];

    for i in 0..n {
        result[i][i] = 0;
        let mut next = VecDeque::new();
        next.push_back(i);
        while let Some(j) = next.pop_front() {
            let d_next = result[i][j] + 1;
            for j in &valves[j].tunnels {
                let d_tunnel = result[i].get_mut(*j).unwrap();
                if d_next < *d_tunnel {
                    *d_tunnel = d_next;
                    next.push_back(*j);
                }
            }
        }
    }

    result
}

#[derive(Debug, Clone)]
pub struct State {
    pub opened: u64,
    pub location: usize,
    pub flow_rate: usize,
    pub time_remaining: usize,
    pub pressure_released: usize,

    interesting: HashSet<usize>,
    history: Vec<usize>,
}

impl State {
    pub fn new(valves: &Vec<Valve>) -> State {
        let interesting = valves
            .iter()
            .enumerate()
            .filter_map(|(i, v)| if v.flow_rate > 0 { Some(i) } else { None })
            .collect();
        State {
            opened: 0,
            location: 0, // "AA" has to be the first valve alphabetically
            flow_rate: 0,
            time_remaining: TOTAL_TIME,
            pressure_released: 0,
            interesting,
            history: vec![0],
        }
    }
}

type ExtraArgs<'a> = (&'a Vec<Valve>, &'a Vec<Vec<usize>>);
impl util::BnBState<ExtraArgs<'_>> for State {
    fn possible_actions(&self, (valves, dists): &ExtraArgs) -> Vec<State> {
        assert!(!self.finished());

        let mut result = Vec::new();

        let loc = self.location;

        if (self.opened & (1 << loc)) == 0 && self.interesting.contains(&loc) {
            let mut open = self.clone();
            open.opened |= 1 << loc;
            // First increase pressure released, then the flow rate
            open.pressure_released += open.flow_rate;
            open.flow_rate += valves[loc].flow_rate;
            open.time_remaining -= 1;
            open.interesting.remove(&loc);
            result.push(open);
        }

        for l in &self.interesting {
            let i = l.clone();
            let dist = dists[loc][i];
            if dist == 0 || dist == usize::MAX || self.time_remaining < dist {
                continue;
            }
            let mut moved = self.clone();
            moved.location = i;
            moved.pressure_released += moved.flow_rate * dist;
            moved.time_remaining -= dist;
            moved.history.push(i);
            result.push(moved);
        }

        let mut do_nothing = self.clone();
        do_nothing.pressure_released += do_nothing.flow_rate * self.time_remaining;
        do_nothing.time_remaining = 0;
        result.push(do_nothing);

        result
    }

    fn finished(&self) -> bool {
        self.time_remaining == 0 || self.interesting.is_empty()
    }

    fn lower_bound(&self, _: &ExtraArgs) -> usize {
        self.pressure_released + self.flow_rate * self.time_remaining
    }

    fn upper_bound(&self, (valves, dists): &ExtraArgs) -> usize {
        let mut additional_flow = 0;
        for i in &self.interesting {
            let i = i.clone();
            if (self.opened & (1 << i)) != 0 {
                continue;
            }
            let dist = dists[self.location][i];
            if dist < self.time_remaining {
                let flow = valves[i].flow_rate;
                additional_flow += flow * (self.time_remaining - dist - 1);
            }
        }

        self.lower_bound(&(valves, dists)) + additional_flow
    }
}

impl Hash for State {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.opened.hash(state);
        self.location.hash(state);
        self.time_remaining.hash(state);
        self.pressure_released.hash(state);
    }
}

impl PartialEq for State {
    fn eq(&self, other: &Self) -> bool {
        self.opened == other.opened
            && self.location == other.location
            && self.time_remaining == other.time_remaining
            && self.pressure_released == other.pressure_released
    }
}

impl Eq for State {}
